 /*
 * sig.c
 * Einfaches Beispiel fuer  Pthreads und Signale.
 */
#include <pthread.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>

#define MAX_NUM_THREADS  2

void *wait_async_sig(void *p)
{
  int caught;
  sigset_t  sigs_to_catch;

  printf("%d: wait_async_sig(): Thread %02d: Sende Signal SIGUSR1(%d) an PID %d\n", 
          (int)getpid(), (unsigned int) pthread_self()%100,SIGUSR1,(int) getpid());
  /* blockiere alle Signale ausser SIGUSR1 und SIGUSR2*/
  while(1) {
     sigemptyset(&sigs_to_catch);
     sigaddset(&sigs_to_catch, SIGUSR1);
     sigaddset(&sigs_to_catch, SIGUSR2);
     sigwait(&sigs_to_catch, &caught);
     printf("%d: wait_async_sig(): Thread %02d: Signal %d aufgetreten\n", 
            (int)getpid(), (unsigned int) pthread_self()%100,caught);
     if ( caught == SIGUSR2 )  break; 
  }
  return(NULL);
}

void sync_err_handler(int sig) 
{
  printf("%d: sync_err_handler(): Thread %02d:synchroner Fehler aufgetreten: Signal %d\n", 
	 (int)getpid(), (unsigned int) pthread_self()%100,sig);
  pthread_exit(0);
}

void *gen_sync_sig(void *p)
{
  int i;
  sigset_t sigs_to_catch;

  /* Aktueller Thread */
  printf("%d: gen_sync_sig(): Thread %02d:  arbeitet\n", 
             (int)getpid(), (unsigned int) pthread_self()%100); 
  /* lasse Signael SIGSEGV und SIGBUS durch  */
  sigemptyset(&sigs_to_catch);
  sigaddset(&sigs_to_catch, SIGSEGV);
  sigaddset(&sigs_to_catch, SIGBUS);
  pthread_sigmask(SIG_UNBLOCK, &sigs_to_catch, NULL);
  /* Erzeuge einen synchronen Fehler */
  sleep(20);
  i=0;
  while (1) { 
    i++; 
    if (i%20 == 0 )  
        printf("%d: gen_sync_sig(): Thread %02d:  Zahler: %4d\n", 
               (int)getpid(), (unsigned int) pthread_self()%100,i); 
    if (i%100 == 0) {
           /* Erzeuge einen synchronen Fehler */
      i = *(int *)p;
    }
  }
  return(NULL); 
}

int main()
{
  int       i;
  pthread_t threads[MAX_NUM_THREADS];
  int       num_threads = 0;
  sigset_t  sigs_to_block;
  struct    sigaction action;

  /* aktueller Thread */
  printf("%d: main(): Thread %02d: arbeitet\n",
            (int) getpid(),(unsigned int) pthread_self()%100); 
  /* Alle Signale blockieren */
  sigfillset(&sigs_to_block);
  pthread_sigmask(SIG_BLOCK, &sigs_to_block, NULL);
  /* Abfangen der Signale SIGSEGV und SIGBUS */
  action.sa_handler=sync_err_handler;
  sigaction(SIGSEGV, &action, NULL);
  sigaction(SIGBUS, &action, NULL);
  /* start Thread mit syncronem Fehler*/
  pthread_create(&threads[num_threads],
		 NULL,
                 gen_sync_sig,
		 NULL);
  printf("%d: main(): Thread %02d: Thread %d erzeugt\n",
            (int) getpid(),(unsigned int) pthread_self()%100,num_threads++); 
  /* erzeuge thread fuer asynchrone Signale */
  pthread_create(&threads[num_threads],
		 NULL,
                 wait_async_sig,
		 NULL);
  printf("%d: main(): Thread %02d: Thread %d erzeugt\n",
            (int) getpid(),(unsigned int) pthread_self()%100,num_threads++); 
  printf("%d: main(): Thread %02d: warten auf das Ende der Threads\n",
            (int) getpid(),(unsigned int) pthread_self()%100); 
  for (i = 0; i < num_threads; i++) {
    pthread_join(threads[i], NULL);
    printf("%d: main(): Thread %02d: Ende von Thread %d\n",
            (int) getpid(),(unsigned int) pthread_self()%100,i); 
  }
  printf("%d: main(): Thread %02d: alle Threads fertig\n",
            (int) getpid(),(unsigned int) pthread_self()%100); 
  exit(0);
}

