/*
 * stat_sigwait.c
 */

#include <pthread.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>

#define MAX_NUM_THREADS  10

pthread_mutex_t stats_lock = PTHREAD_MUTEX_INITIALIZER;
int aktionen=0, wert=0;

void *reporter(void *p)
{
  int eingetroffen;
  sigset_t  warte_signal;

  printf("%d: reporter():  Thread: %02d startet\n",
        (int)getpid(),(unsigned int) pthread_self()%100); 
  /* Signal SIGUSR1 in Maske*/
  sigemptyset(&warte_signal);
  sigaddset(&warte_signal, SIGUSR1);
  for (;;) {
      /* Signal SIGUSR1 freigeben und darauf warten */
     sigwait(&warte_signal, &eingetroffen);
     /* sperren Variable */
     pthread_mutex_lock(&stats_lock);
     printf("%d: reporter(): Thread %02d: Aktionen: %d, Wert: %d\n", 
          (int)getpid(),(unsigned int) pthread_self()%100, aktionen, wert);
     pthread_mutex_unlock(&stats_lock);
  }
}

void *worker_thread( int *p) 
{
  time_t now;

  printf("%d: worker_thread(): Thread %02d: startet, Aktionen: %d, Wert %d, Wait: %d\n", 
     (int)getpid(),(unsigned int) pthread_self()%100,aktionen, wert, (*p)*3); 
  for (;;) {
    sleep((*p)*3);
    now = time(NULL);
    pthread_mutex_lock(&stats_lock);
    wert+=(now%60); /* Arbeiten */
    aktionen++;
    
    sleep(3); 
    printf("%d: worker_thread(): Thread %02d: arbeitet, Aktionen: %d, Wert: %d, Wait: %d\n", 
     (int)getpid(),(unsigned int) pthread_self()%100,aktionen, wert, (*p)*3); 
    pthread_mutex_unlock(&stats_lock);
  }
  /* duerfte nie erreicht werden */
  return NULL;
}

int main()
{
  int       i;
  int       wzeit[MAX_NUM_THREADS];
  pthread_t threads[MAX_NUM_THREADS];
  int       num_threads = 0;
  sigset_t  blockierte_signale;


  printf("%d: main(): Thread %02d: startet. SIGUSR1=%d\n",
          (int)getpid(),(unsigned int) pthread_self()%100,SIGUSR1); 
  /* Signale SIGUSR1 fuer alle Threads blockieren */
  sigemptyset(&blockierte_signale);
  sigaddset(&blockierte_signale, SIGUSR1);
  pthread_sigmask(SIG_BLOCK, &blockierte_signale, NULL);

  pthread_create(&threads[num_threads],
	         NULL,
		 reporter,
		 NULL);
  printf("%d: main(): Thread %02d: Thread %d (reporter) gestartet\n",
          (int)getpid(),(unsigned int) pthread_self()%100,num_threads++); 
  for (i=num_threads; i<MAX_NUM_THREADS; i++) {
    wzeit[i]=i;
    pthread_create(&threads[i],
		   NULL,
                   (void *) worker_thread,
		   &wzeit[i]);
    printf("%d: main(): Thread %02d: Thread %d (worker) gestartet\n",
          (int)getpid(),(unsigned int) pthread_self()%100,num_threads++); 
  }
 
  printf("%d: main(): Thread %02d: alle Threads gestartet\n",
          (int)getpid(),(unsigned int) pthread_self()%100); 
  
  /* auf alle Threads warten */
  for (i = 0; i < num_threads; i++) {
    pthread_join(threads[i], NULL);
    printf("%d: main(): Thread %02d: alle Threads %d beendet\n",
          (int)getpid(),(unsigned int) pthread_self()%100,i); 
  }
  
  printf("%d: main(): Thread %02d: alle Threads (%d) beendet\n",
          (int)getpid(),(unsigned int) pthread_self()%100,i); 
  exit(0);
}

